import construct
from construct import FocusedSeq, stream_tell, Construct, Optional, Const


def Skip(count, subcon):
    return construct.Seek(count * subcon.length, 1)


class LazyPatchedForBug(construct.Lazy):
    r"""
    See https://github.com/construct/construct/issues/938
    """

    def _parse(self, stream, context, path):
        offset = stream_tell(stream, path)

        def execute():
            fallback = stream_tell(stream, path)
            construct.stream_seek(stream, offset, 0, path)
            obj = self.subcon._parsereport(stream, context, path)
            construct.stream_seek(stream, fallback, 0, path)
            return obj

        length = self.subcon._actualsize(stream, context, path)
        construct.stream_seek(stream, length, 1, path)
        return execute


class ErrorWithMessage(Construct):
    def __init__(self, message):
        super().__init__()
        self.message = message
        self.flagbuildnone = True

    def _parse(self, stream, context, path):
        message = construct.evaluate(self.message, context)
        raise construct.ExplicitError(f"Error field was activated during parsing with error {message}", path=path)

    def _build(self, obj, stream, context, path):
        message = construct.evaluate(self.message, context)
        raise construct.ExplicitError(f"Error field was activated during building with error {message}", path=path)

    def _sizeof(self, context, path):
        raise construct.SizeofError("Error does not have size, because it interrupts parsing and building", path=path)


def ForceQuit():
    def force_quit(ctx):
        raise SystemExit(1)
    return ErrorWithMessage(force_quit)


def LabeledOptional(label, subcon):
    return Optional(
        FocusedSeq(
            "subcon",
            Const(label),
            "subcon" / subcon,
        )
    )


def OptionalValue(subcon):
    return construct.FocusedSeq(
        "value",
        present=construct.Rebuild(construct.Flag, construct.this.value != None),
        value=construct.If(construct.this.present, subcon),
    )



from mercury_engine_data_structures import common_types
from mercury_engine_data_structures.common_types import Float


def test_make_dict_parse():
    data = b'\x02\x00\x00\x00foo\x00\x00\x00@?banana\x00\x00 \xa7D'
    con = common_types.make_dict(Float)

    # Run
    r = con.parse(data)

    # Assert
    assert r == {
        "foo": 0.75,
        "banana": 1337,
    }


def test_make_vector():
    data = ["banana", "foobar", "alfafa"]
    con = common_types.make_vector(common_types.StrId)

    # Run
    encoded = con.build(data)
    decoded = con.parse(encoded)

    # Assert
    assert data == decoded




from construct import Adapter, AdaptationError


class OffsetAdapter(Adapter):
    # stores offsets as indices

    def _get_table(self, context):
        raise NotImplementedError()

    def _get_table_length(self, context):
        raise NotImplementedError()

    def _get_base_offset(self, context):
        return 0

    def _get_item_size(self, item):
        return item.size

    def _decode(self, obj, context, path):
        table = self._get_table(context)
        offset = obj
        size = self._get_base_offset(context)

        for i in range(self._get_table_length(context)):
            if size == offset:
                return i
            if size > offset:
                raise AdaptationError("No string begins at the requested offset!")

            item = table[i]
            size += self._get_item_size(item)

    def _encode(self, obj, context, path):
        table = self._get_table(context)
        index = obj
        size = self._get_base_offset(context)

        for i in range(self._get_table_length(context)):
            if i == index:
                return size

            item = table[i]
            size += self._get_item_size(item)



import functools
import json
from pathlib import Path
from typing import Dict, Optional

import construct
from construct import Hex, Int64ul
from construct.lib import HexDisplayedInteger


@functools.lru_cache()
def all_asset_id_to_name() -> Dict[int, str]:
    path = Path(__file__).parent.joinpath("resource_names.json")
    with path.open() as names_file:
        names: Dict[str, int] = json.load(names_file)

    return {
        asset_id: name
        for name, asset_id in names.items()
    }


def name_for_asset_id(asset_id: int) -> Optional[str]:
    return all_asset_id_to_name().get(asset_id)


@functools.lru_cache()
def all_name_to_property_id() -> Dict[str, int]:
    path = Path(__file__).parent.joinpath("property_names.json")
    with path.open() as names_file:
        return json.load(names_file)


@functools.lru_cache()
def all_property_id_to_name() -> Dict[int, str]:
    names = all_name_to_property_id()

    return {
        asset_id: name
        for name, asset_id in names.items()
    }


PropertyEnum = construct.Enum(Hex(Int64ul), **{
    name: HexDisplayedInteger.new(property_id, "0%sX" % (2 * 8))
    for property_id, name in all_property_id_to_name().items()
})



import os
from pathlib import Path

import pytest


def get_env_or_skip(env_name):
    if env_name not in os.environ:
        pytest.skip(f"Skipped due to missing environment variable {env_name}")
    return os.environ[env_name]


@pytest.fixture()
def samus_returns_path():
    return Path(get_env_or_skip("SAMUS_RETURNS_PATH"))


@pytest.fixture()
def dread_path():
    return Path(get_env_or_skip("DREAD_PATH"))




import os
import subprocess
import sys
from pathlib import Path

from setuptools import setup
from setuptools.command.build_py import build_py


def generate_property_templates():
    subprocess.run([
        sys.executable,
        os.fspath(Path(__file__).parent.joinpath("tools", "create_class_definitions.py"))
    ], check=True)


class BuildPyCommand(build_py):
    """
    Generate script templates code before building the package.
    """

    def run(self):
        generate_property_templates()
        build_py.run(self)


setup(
    cmdclass={
        'build_py': BuildPyCommand,
    },
)



import construct

from mercury_engine_data_structures.formats import BaseResource, game_model_root
from mercury_engine_data_structures.game_check import Game

BRSA = game_model_root.create('CSubAreaManager', 0x02010002)


class Brsa(BaseResource):
    @classmethod
    def construct_class(cls, target_game: Game) -> construct.Construct:
        return BRSA



import construct
from construct import Construct, Struct, Const, Hex, Int16ul, Switch, Array, Optional, Rebuild, Flag, Probe, Terminated

from mercury_engine_data_structures.common_types import UInt, make_vector, StrId, Float
from mercury_engine_data_structures.construct_extensions.misc import ErrorWithMessage, ForceQuit, OptionalValue
from mercury_engine_data_structures.formats import BaseResource
from mercury_engine_data_structures.game_check import Game

CollisionPoint = Struct(
    x=Float,
    y=Float,
    material_attribute=UInt,
)

CollisionPoly = Struct(
    num_points=Rebuild(UInt, construct.len_(construct.this.points)),
    unk=Float,
    points=Array(construct.this.num_points, CollisionPoint),
    loop=Flag,
    boundings=Array(4, Float),
)

BinarySearchTree = Struct(
    binary_search_index1=Int16ul,
    binary_search_index2=Int16ul,
    boundings=Array(4, Float),
)

_collision_formats = {
    "CIRCLE": Struct(
        value1=Float,
        value2=Float,
        value3=Float,
        size=Float,
    ),
    "CAPSULE2D": Struct(
        value1=Float,
        value2=Float,
        value3=Float,
        value4=Float,
        value5=Float,
    ),
    "POLYCOLLECTION2D": Struct(
        unknown1=UInt,
        unknown2=UInt,
        unknown3=UInt,
        polys=make_vector(CollisionPoly),
        total_boundings=Array(4, Float),
        binary_search_trees=OptionalValue(make_vector(BinarySearchTree)),
    ),
}

CollisionEntry = Struct(
    name=StrId,
    prop1=StrId,
    prop2=StrId,
    prop3=StrId,
    flag=Int16ul,
    type=StrId,
    data=Switch(
        construct.this.type,
        _collision_formats,
        ErrorWithMessage(
            lambda ctx: f"Type {ctx.type} not known, valid types are {list(_collision_formats.keys())}."
        )
    ),
)

CollisionLayer = Struct(
    name=StrId,
    entries=make_vector(CollisionEntry),
)

BMSCC = Struct(
    magic=Const(0x4443534D, Hex(UInt)),
    version=Const(0x00100001, Hex(UInt)),
    layers=make_vector(CollisionLayer),
    _eof=Terminated,
)


class Bmscc(BaseResource):
    @classmethod
    def construct_class(cls, target_game: Game) -> Construct:
        return BMSCC



import copy
import hashlib
import json
import os
import typing
from pathlib import Path

import numpy

from mercury_engine_data_structures import hashed_names
from mercury_engine_data_structures.formats import Bmscc, Brfld, Brsa
from mercury_engine_data_structures.game_check import Game
from mercury_engine_data_structures.pkg_editor import PkgEditor

area_names = {
    'maps/levels/c10_samus/s010_cave/s010_cave.brfld': "Artaria",
    'maps/levels/c10_samus/s020_magma/s020_magma.brfld': "Cataris",
    'maps/levels/c10_samus/s030_baselab/s030_baselab.brfld': "Dairon",
    'maps/levels/c10_samus/s040_aqua/s040_aqua.brfld': "Burenia",
    'maps/levels/c10_samus/s050_forest/s050_forest.brfld': "Ghavoran",
    'maps/levels/c10_samus/s060_quarantine/s060_quarantine.brfld': "Elun",
    'maps/levels/c10_samus/s070_basesanc/s070_basesanc.brfld': "Ferenia",
    'maps/levels/c10_samus/s080_shipyard/s080_shipyard.brfld': "Hanubia",
    'maps/levels/c10_samus/s090_skybase/s090_skybase.brfld': "Itorash",
}
pickup_index = 0

_polygon_override = {
    ("s010_cave", "collision_camera_010"): [
        [-1200.0, 2200.0],
        [-5200.0, 2200.0],
        [-5200.0, -5600.0],
        [-4800.0, -5600.0],
        [-3300.0, -6200.0],
        [-1200.0, -6200.0]
    ],
    ("s010_cave", "collision_camera_023"): [
        [24900.0, 5100.0], [17800.0, 5100.0], [17800.0, 3500.0], [24900.0, 3500.0]
    ],
    ("s010_cave", "collision_camera_048"): [
        [800.0, 3900.0],
        [-1200.0, 3900.0],
        [-1200.0, 5200.0],
        [-5200.0, 5200.0],
        [-5200.0, 2100.0],
        [-1000.0, 2100.0],
        [-1000.0, 1500.0],
        [500.0, 1500.0],
        [800.0, 1500.0]
    ],
    ("s040_aqua", "collision_camera_023_B"): [
        [4100.0, -9800.0],
        [2000.0, -9800.0],
        [2000.0, -12150.0],
        [3500.0, -12150.0],
        [3500.0, -12500.0],
        [4100.0, -12500.0],
    ],
    ("s070_basesanc", "collision_camera_003"): [
        [-4400.0, -2100.0],
        [-10500.0, -2100.0],
        [-10500.0, -4900.0],
        [-8400.0, -4900.0],
        [-8400.0, -3800.0],
        [-6400.0, -3800.0],
        [-4400.0, -4100.0],
    ],
}
_camera_skip = {
    ("s010_cave", "collision_camera_999"),
    ("s040_aqua", "collision_camera_001_B"),
    ("s050_forest", "collision_camera_024_B"),
    ("s050_forest", "collision_camera_025"),
    ("s050_forest", "collision_camera_025_C"),
    ("s080_shipyard", "collision_camera_009_C"),
}
dock_weakness = {
    "frame": 0,
    "closed": 1,
    "power": 2,
    "charge": 3,
    "presence": 9,
    "grapple": 10,
    "phase_shift": 11,
}
_weakness_table_for_def = {
    'doorframe': (dock_weakness["frame"], dock_weakness["frame"]),
    'doorchargecharge': (dock_weakness["charge"], dock_weakness["charge"]),
    'doorclosedcharge': (dock_weakness["closed"], dock_weakness["charge"]),
    'doorpowerpower': (dock_weakness["power"], dock_weakness["power"]),
    'doorclosedpower': (dock_weakness["closed"], dock_weakness["power"]),
    'doorpowerclosed': (dock_weakness["power"], dock_weakness["closed"]),
    'doorchargeclosed': (dock_weakness["charge"], dock_weakness["closed"]),
    'doorpresencepresence': (dock_weakness["presence"], dock_weakness["presence"]),
    'doorpresenceframe': (dock_weakness["presence"], dock_weakness["frame"]),
    'doorframepresence': (dock_weakness["frame"], dock_weakness["presence"]),
    'doorgrapplegrapple': (dock_weakness["grapple"], dock_weakness["grapple"]),
    'doorclosedgrapple': (dock_weakness["closed"], dock_weakness["grapple"]),
    'doorgrappleclosed': (dock_weakness["grapple"], dock_weakness["closed"]),
    'doorshutter': (dock_weakness["phase_shift"], dock_weakness["phase_shift"]),

    # These doors are going to be event or whatever it is
    'doorheat': (None, None),
}


def decode_world(root: Path, target_level: str):
    global pickup_index
    all_names = hashed_names.all_asset_id_to_name()
    game = Game.DREAD

    with PkgEditor.open_pkgs_at(root) as pkg_editor:
        pkg_editor = typing.cast(PkgEditor, pkg_editor)

        bmscc: Bmscc = None
        brsa: Brsa = None
        brfld: Brfld = None
        brfld_path: str

        for asset_id, name in all_names.items():
            if target_level not in name:
                continue

            if name.endswith("bmscc"):
                print(f"Reading {name}...")
                bmscc = Bmscc.parse(pkg_editor.get_asset_with_asset_id(asset_id), game)

            elif name.endswith("brsa"):
                print(f"Reading {name}...")
                brsa = Brsa.parse(pkg_editor.get_asset_with_asset_id(asset_id), game)

            elif name.endswith("brfld"):
                print(f"Reading {name}...")
                brfld = Brfld.parse(pkg_editor.get_asset_with_asset_id(asset_id), game)
                brfld_path = name

    if bmscc is None or brsa is None:
        raise ValueError("DATA IS NONE")

    cams: dict[str, set[str]] = {}

    for setup in brsa.raw.Root.pSubareaManager.vSubareaSetups:
        for config in setup.vSubareaConfigs:
            for cam in config.vsCameraCollisionsIds:
                cams[cam] = cams.get(cam, set())
                if config.sCharclassGroupId:
                    cams[cam].add(config.sCharclassGroupId)

    import matplotlib.pyplot as plt
    from shapely.geometry import Point
    from shapely.geometry.polygon import Polygon
    from matplotlib.patches import Polygon as mtPolygon

    def rand_color(s):
        return [x / 300.0 for x in hashlib.md5(bytes(str(sorted(s)), 'ascii')).digest()[0:3]]

    handles = []
    rooms = {}
    world: dict[str, typing.Any] = {
        "name": area_names[brfld_path],
        "dark_name": None,
        "asset_id": brfld_path,
        "areas": []
    }

    area_by_name: dict[str, dict] = {}
    plt.figure(1, figsize=(20, 10))
    plt.title(target_level)

    for entry in bmscc.raw.layers[0].entries:
        assert entry.type == "POLYCOLLECTION2D"

        x1, y1, x2, y2 = entry.data.total_boundings
        if abs(x1) > 59999 or abs(y1) > 59999 or abs(x2) > 59999 or abs(y2) > 59999:
            continue

        if (target_level, entry.name) in _camera_skip:
            continue

        if entry.name not in cams:
            continue

        assert len(entry.data.polys) == 1
        raw_vertices = []
        for v in entry.data.polys[0].points:
            raw_vertices.append((v.x, v.y))

        raw_vertices = _polygon_override.get((target_level, entry.name), raw_vertices)
        vertices = numpy.array(raw_vertices)
        c = [0.2, 0.7, 0.6]

        patch = mtPolygon(vertices, linewidth=1, edgecolor=c, facecolor=(c[0], c[1], c[2], 0.1))
        plt.gca().add_patch(patch)
        plt.text((x1 + x2) / 2, (y1 + y2) / 2, entry.name[17:], color=c, ha='center', size='x-small')
        handles.append(patch)
        rooms[entry.name] = Polygon(vertices)
        world["areas"].append({
            "name": f"{entry.name} ({area_names[brfld_path][0]})",
            "in_dark_aether": False,
            "asset_id": entry.name,
            "default_node_index": None,
            "valid_starting_location": False,
            "nodes": [],
            "extra": {
                "total_boundings": {
                    "x1": x1,
                    "x2": x2,
                    "y1": y1,
                    "y2": y2,
                },
                "polygon": raw_vertices,
            }
        })
        area_by_name[entry.name] = world["areas"][-1]

    handles_by_label = {}
    door_color = [0.8, 0.2, 0.2]
    item_color = [0.2, 0.2, 0.8]
    actor_positions = {}

    def _find_room_orientation(room_a: str, room_b: str):
        a_bounds = area_by_name[room_a]["extra"]["total_boundings"]
        b_bounds = area_by_name[room_b]["extra"]["total_boundings"]
        a_cx = (a_bounds["x1"] + a_bounds["x2"]) / 2
        b_cx = (b_bounds["x1"] + b_bounds["x2"]) / 2
        if a_cx < b_cx:
            return 0, 1
        elif a_cx > b_cx:
            return 1, 0
        else:
            raise ValueError(f"{room_a} and {room_b} are aligned")

    def count_docks(rm: str) -> int:
        return sum(
            1 for node in area_by_name[rm]["nodes"]
            if node["node_type"] == "dock"
        )

    def get_def_link_for_entity(actor_ref: str) -> typing.Optional[str]:
        a = brfld.follow_link(actor_ref)
        if a is not None:
            return a.oActorDefLink

    for actor in brfld.actors_for_layer("default").values():
        actor_def = os.path.splitext(os.path.split(actor.oActorDefLink)[1])[0]
        # is_door = "door" in actor.sName.lower()
        is_door = "LIFE" in actor.pComponents and "CDoorLifeComponent" == actor.pComponents.LIFE["@type"]

        if is_door or "actors/items" in actor.oActorDefLink:
            handles_by_label[actor_def], = plt.plot(actor.vPos[0], actor.vPos[1], "o", color=rand_color(actor_def))

            p = Point(actor.vPos)
            if others := [name for name, other in actor_positions.items() if p.distance(other) < 3]:
                print(f"{actor.sName} is very close to {[other for other in others]}")
            actor_positions[actor.sName] = p
            rooms_for_actor: list[str] = [name for name, pol in rooms.items() if pol.contains(p)]

            va = "bottom" if is_door else "top"
            plt.annotate(actor.sName, actor.vPos[:2], fontsize='xx-small', ha='center', va=va)
            if is_door:
                extra = {
                    "actor_def": actor.oActorDefLink,
                    "left_shield_entity": actor.pComponents.LIFE.wpLeftDoorShieldEntity,
                    "left_shield_def": get_def_link_for_entity(actor.pComponents.LIFE.wpLeftDoorShieldEntity),
                    "right_shield_entity": actor.pComponents.LIFE.wpRightDoorShieldEntity,
                    "right_shield_def": get_def_link_for_entity(actor.pComponents.LIFE.wpRightDoorShieldEntity),
                }

                if len(rooms_for_actor) == 2:
                    simple = {"def": actor_def,
                              "left": extra["left_shield_def"],
                              "right": extra["right_shield_def"]}

                    left_room, right_room = _find_room_orientation(*rooms_for_actor)
                    custom_weakness = [None, None]

                    if simple["left"] == simple["right"] and simple["left"] is None:
                        if actor_def in _weakness_table_for_def:
                            custom_weakness[left_room], custom_weakness[right_room] = _weakness_table_for_def[actor_def]
                        else:
                            print(f"no weakness for {actor_def} without shields")

                    door_template = {
                        "name": f"Door ({actor.sName})",
                        "heal": False,
                        "coordinates": {
                            "x": actor.vPos[0],
                            "y": actor.vPos[1],
                            "z": actor.vPos[2],
                        },
                        "extra": extra,
                        "node_type": "dock",
                        "dock_index": None,
                        "connected_area_asset_id": None,
                        "connected_dock_index": None,
                        "dock_type": 2,
                        "dock_weakness_index": 1,
                        "connections": {}
                    }

                    doors = []
                    for i, room_name in enumerate(rooms_for_actor):
                        door = copy.copy(door_template)
                        door["dock_index"] = count_docks(room_name)
                        door["connected_area_asset_id"] = rooms_for_actor[(i + 1) % 2]
                        door["connected_dock_index"] = count_docks(rooms_for_actor[(i + 1) % 2])
                        if custom_weakness[i] is not None:
                            door["dock_type"] = 0
                            door["dock_weakness_index"] = custom_weakness[i]
                        doors.append(door)

                    for i, room_name in enumerate(rooms_for_actor):
                        area_by_name[room_name]["nodes"].append(doors[i])

                elif len(rooms_for_actor) > 2:
                    print("multiple rooms for door!", actor.sName, rooms_for_actor)
            else:
                for room_name in rooms_for_actor:
                    area_by_name[room_name]["nodes"].append({
                        "name": f"Pickup ({actor.sName})",
                        "heal": False,
                        "coordinates": {
                            "x": actor.vPos[0],
                            "y": actor.vPos[1],
                            "z": actor.vPos[2],
                        },
                        "extra": {
                            "actor_def": actor.oActorDefLink,
                        },
                        "node_type": "pickup",
                        "pickup_index": pickup_index,
                        "major_location": "tank" not in actor_def,
                        "connections": {}
                    })
                if len(rooms_for_actor) != 1:
                    print("wrong item!", actor.sName, rooms_for_actor)
                pickup_index += 1

        # else:
        #     plt.annotate(actor.sName, actor.vPos[:2], fontsize='xx-small', ha='center')
        #     plt.plot(actor.vPos[0], actor.vPos[1], "o", color=item_color)

    handles_by_label = {
        key: value
        for key, value in sorted(handles_by_label.items(), key=lambda it: it[0])
    }

    for area in world["areas"]:
        if not area["nodes"]:
            area["nodes"].append({
                "name": "Placeholder",
                "heal": False,
                "coordinates": None,
                "extra": {},
                "node_type": "generic",
                "connections": {},
            })

    plt.legend(handles_by_label.values(), handles_by_label.keys())
    # plt.show()
    # plt.savefig(f"{target_level}.png", dpi=200, bbox_inches='tight')
    plt.close()

    with open(f"{world['name']}.json", "w") as f:
        json.dump(world, f, indent=4)


def decode_all_worlds(root: Path):
    for area_path in area_names.keys():
        level_name = os.path.splitext(os.path.split(area_path)[1])[0]
        decode_world(root, level_name)


if __name__ == '__main__':
    decode_all_worlds(Path("E:/DreadExtract"))
    # decode_world(Path("E:/DreadExtract"), "s010_cave")



import construct

from mercury_engine_data_structures.formats import BaseResource, game_model_root
from mercury_engine_data_structures.game_check import Game

BRFLD = game_model_root.create('CScenario', 0x02000031)


class Brfld(BaseResource):
    @classmethod
    def construct_class(cls, target_game: Game) -> construct.Construct:
        return BRFLD

    def actors_for_layer(self, name: str) -> dict:
        return self.raw.Root.pScenario.rEntitiesLayer.dctSublayers[name].dctActors

    def all_actors(self):
        for sublayer in self.raw.Root.pScenario.rEntitiesLayer.dctSublayers.values():
            yield from sublayer.dctActors.values()

    def follow_link(self, link: str):
        if link != '{EMPTY}':
            result = self.raw
            for part in link.split(":"):
                result = result[part]
            return result



import contextlib
from contextlib import ExitStack
from pathlib import Path
from typing import BinaryIO, Dict, Optional, Generator, Iterator

from mercury_engine_data_structures import crc
from mercury_engine_data_structures.formats.base_resource import AssetId
from mercury_engine_data_structures.formats.pkg import PKGHeader
from mercury_engine_data_structures.game_check import Game


class PkgEditor:
    """
    Manages efficiently reading all PKGs in the game and writing out modifications to a new path.
    """

    def __init__(self, files: Dict[str, BinaryIO]):
        self.files = files
        self.headers = {
            name: PKGHeader.parse_stream(file, target_game=Game.DREAD)
            for name, file in files.items()
        }

    @classmethod
    @contextlib.contextmanager
    def open_pkgs_at(cls, root: Path) -> Generator["PkgEditor", None, None]:
        all_pkgs = root.rglob("*.pkg")

        with ExitStack() as stack:
            files = {
                file.relative_to(root).as_posix(): stack.enter_context(file.open("rb"))
                for file in all_pkgs
            }
            yield PkgEditor(files)

    def find_pkgs_for_asset_id(self, asset_id: AssetId) -> Iterator[str]:
        for name, header in self.headers.items():
            if any(entry.asset_id == asset_id for entry in header.file_entries):
                yield name

    def find_pkgs_for_name(self, name: str) -> Iterator[str]:
        yield from self.find_pkgs_for_asset_id(crc.crc64(name.encode("utf-8")))

    def get_asset_with_asset_id(self, asset_id: AssetId, in_pkg: Optional[str] = None) -> bytes:
        for name, header in self.headers.items():
            if in_pkg is not None and name != in_pkg:
                continue

            for entry in header.file_entries:
                if entry.asset_id == asset_id:
                    self.files[name].seek(entry.start_offset)
                    return self.files[name].read(entry.end_offset - entry.start_offset)
        raise ValueError(f"Unknown asset_id: {asset_id:0x}")

    def all_asset_ids(self):
        return {
            entry.asset_id
            for header in self.headers.values()
            for entry in header.file_entries
        }



import collections
import json
import re
from pathlib import Path
from typing import Optional

known_types_to_construct = {
    "bool": "construct.Flag",
    "float": "common_types.Float",
    "float32": "common_types.Float",
    "int": "common_types.Int",
    "unsigned": "common_types.UInt",
    "unsigned_int": "common_types.UInt",
    "base::global::CStrId": "common_types.StrId",
    "base::global::CFilePathStrId": "common_types.StrId",
    "base::global::CRntString": "common_types.StrId",
    "base::math::CVector2D": "common_types.CVector2D",
    "base::math::CVector3D": "common_types.CVector3D",
    "base::math::CVector4D": "common_types.CVector4D",
    "CGameLink<CActor>": "common_types.StrId",
    "CGameLink<CEntity>": "common_types.StrId",
    "CGameLink<CSpawnPointComponent>": "common_types.StrId",
    "base::global::CRntFile": "construct.Prefixed(construct.Int32ul, construct.GreedyBytes)",

    # ESubAreaItem::Count = 9
    "base::global::CArray<base::global::CStrId, EnumClass<ESubAreaItem>::Count, ESubAreaItem>": (
        "common_types.make_vector(common_types.StrId)"
    ),

    # TODO: test if works
    "base::global::CName": "common_types.StrId",
    "base::core::CAssetLink": "common_types.StrId",

    # Hacky
    # Enums that don't use enum-like names in fields
    "CDoorLifeComponent::SState": "common_types.UInt",
    "base::snd::ELowPassFilter": "common_types.UInt",
    "EShinesparkTravellingDirection": "common_types.UInt",
    "ECoolShinesparkSituation": "common_types.UInt",
}

vector_re = re.compile(r"(?:base::)?global::CRntVector<(.*?)(?:, false)?>$")
dict_re = re.compile(r"base::global::CRnt(?:Small)?Dictionary<base::global::CStrId,[\s_](.*)>$")
all_container_re = {
    "common_types.make_vector": vector_re,
    "common_types.make_dict": dict_re,
}

unique_ptr_re = re.compile(r"std::unique_ptr<(.*)>$")
weak_ptr_re = re.compile(r"base::global::CWeakPtr<(.*)>$")
raw_ptr_re = re.compile(r"(.*?)(?:[ ]?const)?\*$")
ref_re = re.compile(r"CGameObjectRef<(.*)>$")
typed_var_re = re.compile(r"(base::reflection::CTypedValue)$")
all_ptr_re = [unique_ptr_re, weak_ptr_re, raw_ptr_re, ref_re, typed_var_re]


def _type_name_to_python_identifier(type_name: str):
    return type_name.replace("::", "_").replace(" ", "_").replace("<", "_").replace(
        ">", "_").replace(",", "_").replace("*", "Ptr")


class TypeExporter:
    def __init__(self, all_types: dict[str, dict[str, str]]):
        self.all_types = all_types
        self._exported_types = {}
        self._types_with_pointer = set()
        self._types_being_exported = set()
        self._children_for = collections.defaultdict(set)
        self._type_definition_code = ""

        self._exported_types["base::reflection::CTypedValue"] = _type_name_to_python_identifier(
            "base::reflection::CTypedValue"
        )
        self._children_for["base::reflection::CTypedValue"].add("base::global::CRntFile")

        for type_name, data in all_types.items():
            if data["parent"] is not None:
                self._children_for[data["parent"]].add(type_name)

    def children_for(self, type_name: str, recursive: bool = True):
        for child in self._children_for[type_name]:
            yield child
            if recursive:
                yield from self.children_for(child)

    def _debug(self, msg: str):
        print("  " * len(self._types_being_exported) + f"* {msg}")

    def _export_known_type(self, type_variable: str, type_name: str):
        data = self.all_types[type_name]

        parent_name = None
        if data["parent"] is not None:
            parent_name = self.ensure_exported_type(data["parent"])

        if data["fields"]:
            field_lines = []
            for field_name, field_type in data["fields"].items():
                self._debug(f"Exporting field! {field_name} = {field_type}")
                converted_type = self.convert_type_to_construct(field_name, field_type)
                field_lines.append(f'    "{field_name}": {converted_type},')

            if parent_name is not None:
                field_lines.insert(0, f"    **{parent_name}Fields,")

            fields_def = "{{\n{}\n}}".format("\n".join(field_lines))
        elif parent_name is not None:
            # No fields, just use the parent dict
            fields_def = f"{parent_name}Fields"
        else:
            # No fields and no parent, empty dict!
            fields_def = "{}"

        field_var = ""
        if self._children_for[type_name]:
            # We have children, create a field vars
            field_var = f"{type_variable}Fields := "

        return f'Object({field_var}{fields_def})'

    def _export_type(self, type_name: str):
        type_variable = _type_name_to_python_identifier(type_name)

        if type_name in self.all_types:
            type_code = self._export_known_type(type_variable, type_name)
        else:
            type_code = self.convert_type_to_construct(None, type_name)

        self._type_definition_code += f'\n\n{type_variable} = {type_code}'
        self._exported_types[type_name] = type_variable

    def ensure_exported_type(self, type_name: str) -> str:
        if type_name in self._exported_types:
            return self._exported_types[type_name]

        self._debug(f"Exporting new type! {type_name}")

        if type_name in self._types_being_exported:
            raise RuntimeError(f"Recursive export for {type_name} detected.")

        self._types_being_exported.add(type_name)
        self._export_type(type_name)
        self._types_being_exported.remove(type_name)

        self._debug(f"Finish exporting type {type_name}")

        return self._exported_types[type_name]

    def pointer_to_type(self, type_name: str) -> str:
        self._types_with_pointer.add(type_name)
        self.ensure_exported_type(type_name)
        return "Pointer_" + _type_name_to_python_identifier(type_name)

    def convert_type_to_construct(self, field_name: Optional[str], field_type: str):
        if field_type in known_types_to_construct:
            return known_types_to_construct[field_type]

        if field_name is not None and field_name.startswith("e"):
            return "common_types.UInt"

        # Containers
        try:
            make, m = next((make, x) for make, r in all_container_re.items() if (x := r.match(field_type)))
            if (inner_field := self.convert_type_to_construct(field_name, m.group(1))) is not None:
                self._debug(f"Container! {field_name} -> {field_type} -> {make} -> {inner_field}")
                return f"{make}({inner_field})"
            return None
        except StopIteration:
            pass

        # Pointers
        try:
            m = next(x for r in all_ptr_re if (x := r.match(field_type)))
            return f'{self.pointer_to_type(m.group(1))}.create_construct()'
        except StopIteration:
            pass

        if field_type not in self.all_types:
            raise ValueError(f"Unknown type: {field_type}")

        return self.ensure_exported_type(field_type)

    def export_code(self):
        code = """# This file was generated!
import construct

from mercury_engine_data_structures import common_types
from mercury_engine_data_structures.object import Object
from mercury_engine_data_structures.pointer_set import PointerSet

"""
        seen_types_with_pointer = set()
        while unchecked_types := self._types_with_pointer - seen_types_with_pointer:
            for type_name in sorted(unchecked_types):
                seen_types_with_pointer.add(type_name)
                for child in sorted(self.children_for(type_name)):
                    self.ensure_exported_type(child)

        for type_name in sorted(self._types_with_pointer):
            code += '{} = PointerSet("{}")\n'.format(
                self.pointer_to_type(type_name),
                type_name,
            )

        # for type_name in self.all_types.keys():
        #     self.ensure_exported_type(type_name)

        code += self._type_definition_code

        code += "\n\n"

        for type_name in sorted(self._types_with_pointer):
            if type_name != "base::reflection::CTypedValue":
                code += '{}.add_option("{}", {})\n'.format(
                    self.pointer_to_type(type_name),
                    type_name,
                    self.ensure_exported_type(type_name),
                )
            for child in sorted(self.children_for(type_name)):
                code += f'{self.pointer_to_type(type_name)}.add_option("{child}", {self.ensure_exported_type(child)})\n'
            code += "\n"

        return code


def main():
    p = Path(__file__).parent.joinpath("all_types.json")
    output_path = Path(__file__).parents[1].joinpath("mercury_engine_data_structures", "formats", "dread_types.py")

    with p.open() as f:
        all_types: dict[str, dict[str, str]] = json.load(f)

    all_types.pop("CBlackboard")
    all_types.pop("CGameBlackboard")
    all_types["gameeditor::CGameModelRoot"]["fields"].pop("pSoundManager")
    all_types["gameeditor::CGameModelRoot"]["fields"].pop("pShotManager")
    all_types["gameeditor::CGameModelRoot"]["fields"].pop("pLightManager")
    all_types["gameeditor::CGameModelRoot"]["fields"].pop("pMusicManager")

    type_exporter = TypeExporter(all_types)
    type_exporter.ensure_exported_type("gameeditor::CGameModelRoot")

    # needs_exporting = {"CActor"}
    # while needs_exporting:
    #     next_type = needs_exporting.pop()
    #     if next_type not in type_exporter._exported_types:
    #         type_exporter.ensure_exported_type(next_type)
    #         needs_exporting.update(type_exporter._children_for[next_type])

    output_path.write_text(type_exporter.export_code())


if __name__ == '__main__':
    main()
